// Completion.cpp: implementation of the CCompletion class.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Colors.h"
#include "Cubes.h"
#include "Completion.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/////////////////////////////////////////////////////////////////////////////
// From the current state, determines the different configurations possible
// for the Rubik's Cube
/////////////////////////////////////////////////////////////////////////////

typedef enum {                            // errors code
   er_missColor=-1, er_complete=-2, er_tooColors=-3, er_badColor=-4,
   er_invalidColors=-5, er_unsolvable=-6, er_badRotation=-7,
   er_centersUndefine=-8, er_badPermut=-9, er_last=-10
};

static const UINT maxTests = 1000;        // limit of search process
static const UINT maxNodeTbl = 20;
 
static t_CUBIES cubesEdgCrn = 0;          // bits of edge and corner cubies
static BYTE defColorsId[6];               // list of center facet numbers

/////////////////////////////////////////////////////////////////////////////
// Construction/Destruction
/////////////////////////////////////////////////////////////////////////////

CCompletion::CCompletion() : m_edges(2), m_corners(3)
{
   if ( cubesEdgCrn == 0 ) {
      // builds the orientation facets table with the default colors
      BYTE facet, color, facets[54];

      SetSideText(CCubes().GetText(), facets, "ORGBWYUDLRFB");
      for (facet = 4; facet < 54; facet += 9) {
         color = facets[facet];
         ASSERT(color < 6);
         defColorsId[color] = facet;
      }

      // the cubies selected with this class
      cubesEdgCrn = m_edges.m_cubesBits | m_corners.m_cubesBits;
      ASSERT((GetCubiesVSC(2)|GetCubiesVSC(3)) == cubesEdgCrn);
   }
}

CCompletion::~CCompletion()
{
}

/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
//                    Prends en compte un jeu de facettes                  //
//                                                                         //
void CCompletion::Get(const BYTE facets[54])                               //
//=========================================================================//
{
   m_edges.Create(facets);
   m_corners.Create(facets);
}

//=========================================================================//
//                                                                         //
//   Assure la mise a jour des facettes qui n'ont qu'une couleur possible  //
//                                                                         //
void CCompletion::Update(BYTE facets[54]) //
//=========================================================================//
{
   CentersUpdate(facets);
   Get(facets);
   m_edges.Update(facets, m_corners.getFlags());
   m_corners.Update(facets, m_edges.getFlags());
}

//=========================================================================//
//                                                                         //
//        Cherche les couleurs possibles pour une facette donnee           //
//                                                                         //
t_COLORS CCompletion::GetPossibleColors(// retourne les couleurs possibles //
                      BYTE facet,       // no facette a renseigner         //
            const BYTE facets[54]) const// couleurs des facettes           //
//=========================================================================//
{
   t_COLORS colors;
   BYTE type;

   if ((facets[facet] & 0x0f) < 6) {
      // facet already filled

      colors = 0;

   } else if ((type =  getFacetType(facet)) == 1) {
      // facet center of face

      colors = CubeOrientation(facet, facets);

   } else if (type == 2) {
      // facet of edge

      ASSERT(m_edges.Compare(facets));
      ASSERT(m_corners.Compare(facets));
      colors = m_edges.getPossibleColors(facet, m_corners.getFlags());

   } else if (type == 3) {
      // facet of corner

      ASSERT(m_edges.Compare(facets));
      ASSERT(m_corners.Compare(facets));
      colors = m_corners.getPossibleColors(facet, m_edges.getFlags());
   }

   return colors;
}

//=========================================================================//
//                                                                         //
//                     Mets a jour une facette donnee                      //
//                                                                         //
void CCompletion::SetAt(BYTE facet,       // facet number to update 0..53  //
                        BYTE color,       // color to update 0..5 or 0xff  //
                        BYTE facets[54],  // facets table                  //
                        bool upd)         // enable to auto update         //
//=========================================================================//
{
   // update the facet
   if (facet % 9 != 4)   facets[facet] = color;
   // retains the orientation of the face centers
   else facets[facet] = (facets[facet] & 0xf0) | (color & 0x0f);

   Get(facets);
   if (upd) Update(facets);
}

//=========================================================================//
//                                                                         //
//                    Efface toutes les facettes liees                     //
//                                                                         //
void CCompletion::DelLinked(BYTE facet,   // facet number to update        //
                        BYTE facets[54])  // facets table                  //
//=========================================================================//
{
   BYTE side;

   if (getFacetType(facet)== 1) {
      // facets center of face

      for (side = 4; side < 54; side += 9) facets[side] |= 0x0f;

   } else {
      // facets edges or corners

      SetSideColors( GetAllSides(facet), GROUP_EMPTY, facets );
   }
   Get(facets);
}

//=========================================================================//
//                                                                         //
//                 Mets a jour toutes les facettes d'une face              //
//                                                                         //
BYTE CCompletion::FillFullSide(  // return the count of the updated facets //
                        BYTE side,        // to get a color of the side    //
                        BYTE facets[54],  // facets table                  //
                        bool upd)         // enable to update              //
//=========================================================================//
{
   BYTE cur = (side / 9)* 9, color = facets[side] & 0x0f, count = 0, n;

   if (color >= 6) {
      // delete a side

      for (n = 0 ; n < 9; ++n) {
         if ((facets[cur + n] & 0x0f)== 0x0f) continue;
         if (upd) facets[cur + n] |= (n == 4)? 0x0f: 0xff;
         ++count;
      }
      if (upd && count > 0) Get(facets);

   } else {
      t_COLORS cvrt = 1 << color;
      // fill a side

      // first center of side
      if (upd && (GetPossibleColors(cur + 4, facets) & cvrt))
         SetAt(cur + 4, color, facets, true), ++count;

      for (n = 0 ; n < 9; ++n) {
         if (n == 4 || (GetPossibleColors(cur + n, facets)& cvrt)== 0) continue;
         if (upd) SetAt(cur + n, color, facets, true);
         ++count;
      }
   }

   return count;
}

//=========================================================================//
//                                                                         //
//               Mise a jour des couleurs des centres de faces             //
//                                                                         //
int CCompletion::CentersUpdate(BYTE facets[54]) const                      //
//=========================================================================//
{
   BYTE side, count = 0;
   t_COLORS colors;

   for (side = 4; side < 54; side += 9) {
      if ((facets[side] & 0x0f) >= 6) {
         colors = CubeOrientation(side, facets);
         if (colorsCount(colors) == 1) {

            facets[side] = (facets[side] & 0xf0) | colorCvrt(colors);
            ++count;
         }
      }
   }
   return count;
}

//=========================================================================//
//                                                                         //
//             Definition of the color facets, center of faces,            //
//                and the orientation of the Rubik's Cube                  //
//                                                                         //
t_COLORS CCompletion::CubeOrientation(    // return the possible color     //
                           BYTE side,     // box number of a center face   //
            const BYTE facets[54]) const  // 
//=========================================================================//
{
   ASSERT(side % 9 == 4);               // this is not a center of face
   // Search the facets, face center, already filled
   t_GROUP sidesNo = GetOrientSides(facets);

   if (sidesNo == GROUP_EMPTY) {
      // the filled facets can not define an orientation
      BYTE color = facets[oppositeSide(side)] & 0x0f;

      // the opposite side is define, return the opposite color of the side
      if (color < 6) return cvrtColor(color ^ 1);
      else return OrientationColors(side, facets);
   }

   // the orientation is define, extract the corresponding colors,  
   t_GROUP refColors = RefColors(sidesNo);
   t_GROUP colors = GetSideColors(sidesNo, facets);
   // Calculates the corresponding orientation of the Rubik's Cube
   t_Matrix or = GetCubeOrientation(refColors, colors, 0);

   if (or == 0) {
      return 0x80;   // error, bad colors !
   }
   // Calculates the  missing colors of the central facets  by
   // positioning a Rubik's Cube with this orientation

   BYTE tmpFacets[54];
   CString cubeMsg = CCubes().Fixe(0, true, or).GetText();

   SetSideText(cubeMsg, tmpFacets, "ORGBWYUDLRFB");
   
   return cvrtColor(tmpFacets[side]);
}

//=========================================================================//
// the colors present in the centers of faces can not define an orientation//
// of the Cube                                                             //
t_COLORS CCompletion::OrientationColors(BYTE side,    //
                        const BYTE facets[54]) const  //
//=========================================================================//
{
   t_COLORS cur, cvrt = 0;
   BYTE color, n;

   // keeps the colors detected and their complements
   for (n = 4; n < 54; n += 9) {
      if ((color = facets[n] & 0x0f) < 6) cvrt |= 3 << (color & 6);
   }
   // computs the free colors
   cvrt ^= 0x3f;

   // removes the colors that do not provide parity permutations if the edges
   // and corners are completely filled and the new color should help fix the
   // orientation of the Cube
   if ( m_edges.m_testPermutation && colorsCount(cvrt) == 4
     && cubesEdgCrn == (GetCompleteCubies(facets) & cubesEdgCrn) ) {
      CSlv edges(2), corns(3);
      BYTE temp[54];

      ASSERT(side % 9 == 4);

      memcpy(temp, facets, 54);
      for (color = 0, cur = 1; cvrt >= cur; ++color, cur <<= 1) {
         if ( (cvrt & cur)== 0 ) continue;

         // init the colors for the center faces
         for (n = 4; n < 54; n += 9) temp[n] = facets[n] & 0x0f;
         temp[side] = color;

         t_GROUP sidesNo = GetOrientSides(temp);
         t_GROUP refColors = RefColors(sidesNo);
         t_GROUP colors = GetSideColors(sidesNo, temp);
         // Calculates the corresponding orientation of the Rubik's Cube
         t_Matrix or = GetCubeOrientation(refColors, colors, 0);

         ASSERT(or != 0);

         // Calculates the  missing colors of the central facets  by
         // positioning a Rubik's Cube with this orientation

         BYTE tmpFacets[54];
         CString cubeMsg = CCubes().GetText(or);

         SetSideText(cubeMsg, tmpFacets, "ORGBWYUDLRFB");
         for (n = 4; n < 54; n += 9) temp[n] = tmpFacets[n];

         ASSERT(temp[side] == color);

         if (edges.Create(temp)!= er_complete
            || corns.Create(temp)!= er_complete) {
            TRACE("CCompletion::OrientationColors of edges or corners bad\n");
            break;
         }
         // removes colors that do not provide parity permutations
         if ( (edges.getFlags() & corns.getFlags()) == 0 ) {
            cvrt ^= cur;
         }
      }
   }
   return cvrt;
}

//=========================================================================//
//                                                                         //
//             Donne le commentaire correspondant a la facette             //
//                                                                         //
CString CCompletion::GetPossibleMsg(                                       //
                      BYTE facet,               // no facet to update      //
                  const BYTE facets[54]) const  // facets of Rubik's Cube  //
//=========================================================================//
{
   static const TCHAR *sidesTypesMsg[] = {
      "xx","Facette centre de face","Facette arte","Facette sommet"
   };
   const TCHAR *msgFacet = sidesTypesMsg[getFacetType(facet)];
   BYTE cnt = colorsCount(GetPossibleColors(facet, facets));
   CString msg;

   if (cnt == 0)
      msg = msgFacet;

   else if (cnt == 1)
      msg.Format("%s, 1 couleur possible", msgFacet);

   else
      msg.Format("%s, %d couleurs possibles", msgFacet, cnt);
   
   return msg;
}

//=========================================================================//
CString CCompletion::getErrorMsg() const                                   //
//=========================================================================//
{
   int err;

   if ( (err = m_edges.getStatus(m_corners.getFlags()))<= er_centersUndefine
     || (err = m_corners.getStatus(m_edges.getFlags()))<= er_centersUndefine ) {
      return m_edges.getErrorString(err, 0);

   } else {
      return CenterSideMsg() +
         m_edges.getErrorString(m_corners.getFlags()) +
         m_corners.getErrorString(m_edges.getFlags());
   }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Construction/Destruction CCompletion::CSlv
/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
CCompletion::CSlv::CSlv(BYTE nbc)   // type of cubies (2 or 3 facets)      //
//=========================================================================//
{
   BYTE cube, n, org;

   // for all cubies of the Rubik's Cube
   memset(m_reversSides, 0xff, 54);
   memset(m_refCvrt, 0xff, 64);

   ASSERT( nbc == 2 || nbc == 3 );

   m_nbc = nbc;
   m_cubesBits = 0;
   for ( n = org = cube = 0; cube < 27; cube++ ) {
      t_GROUP noSides = getNoSides(cube);
      if ( SidesCount(noSides) != nbc ) continue;

      //  selected only the cubies with a good account of visible facets
      t_GROUP noColors = RefColors(noSides);
      t_COLORS colors = cvrtColors(noColors);
      ASSERT(colors < (1 << 6));    // colors signature of the cubie

      m_cubesBits |= 1 << cube;     // set the selected cubies

      // for all visible facets in the cubie
      // whatever the chosen type, there are in total 24 visible facets
      for (BYTE i = 0; i < nbc; ++i) {
         BYTE facet = BYTE_N(noSides, i), side = org + i;
         ASSERT(side < 24);         // side number in m_work table
         ASSERT(facet < 54);        // facet number in facets table

         // the translation between the facet number and the side table no
         m_convSides[side] = facet;
         m_reversSides[facet] = side;
         // the first side of each cubie in sides table
         m_orgSides[side] = org;
      }

      // the translation between the color signature and the cube No
      m_refCvrt[colors] = cube;
      // the library of color signatures for cubies of this type
      m_refColors[n] = colors;
      ++n, org += nbc;
   }
   // set the end of table of signatures
   m_refColors[n] = 0;
   ASSERT(org == 24);

   // set empty
   m_sCnt = er_tooColors;
   m_flags = 3;
   m_computed = false;
   m_testOrientation = true;
   m_testPermutation = true;
   memset(m_work, 0xff, 24);
   memset(m_solutions, 0, 48);
}

//=========================================================================//
CCompletion::CSlv::~CSlv()                                                 //
//=========================================================================//
{
}

/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
//       Copy the facets to process and initialize the working buffer      //
//                                                                         //
int CCompletion::CSlv::Create(const BYTE facets[54])                       //
//=========================================================================//
{
   BYTE n, c, color, reversId[6];
   bool newData;

   // Transcoding colors facets of Rubik's Cube using color centers face
   // so to have a Rubik's Cube in a standard orientation. This easy to
   // calculate the possible displacements of cubies. 
   for (m_colorsCoding = true, n = 0; n < 6; ++n) {
      if ( (color = facets[defColorsId[n]] & 0x0f) >= 6 ) {
         // until the color of the center of the faces are not defined,
         // there is no color transcoding.
         for (m_colorsCoding = false, n = 0; n < 6; ++n) {
            m_colorsId[n] = 1 << n;
            m_colorId[n] = n;
            reversId[n] = n;
         }
         break;
      }
      m_colorsId[n] = 1 << color;
      m_colorId[n] = color;
      reversId[color] = n;
   }

   // copy the colors and test for a new configuration
   for (newData = false, n = 0; n < 24; ++n) {
      c = facets[ m_convSides[n] ];
      if ( m_work[n] == ((c < 6)? reversId[c]: 0xff) ) continue;

      // copy the color
      for (; n < 24; ++n) {
         c = facets[ m_convSides[n] ];
         m_work[n] = (c < 6)? reversId[c]: 0xff;
      }
      newData = true;
      break;
   }

   // avoid calculating the possible colors if the configuration has not
   // changed.
   return (newData)? Computs(): m_sCnt;
//   return 0;
}

//=========================================================================//
//                                                                         //
//             Test that the configuration matches the stored              //
//                                                                         //
bool CCompletion::CSlv::Compare(const BYTE facets[54]) const               //
//=========================================================================//
{
   BYTE n, c, color;

   for (n = 0; n < 24; ++n) {
      c = m_work[n];
      color = (c < 6)? m_colorId[c]: 0xff;
      if (facets[ m_convSides[n] ] != color)
         return false;
   }
   return true;
}

//=========================================================================//
//           Test the consistency of the colors in the work buffer         //
BYTE CCompletion::CSlv::ValidityCheck() const                              //
//=========================================================================//
{
   BYTE cubes[12], cube, maxCubies = 24 / m_nbc, sumRot;
   t_CUBIES cubesBits;

   // recognize the different cubies (orientation and permutation)
   for (cubesBits = 0, sumRot = cube = 0; cube < maxCubies; ++cube) {
      // calculate the rotation of the sides
      char rot = CheckSidesColors(getColorsNo(cube*m_nbc), cubes+cube);
      if (rot < 0) {
         // error, invalid colors combination
         return 3 | ((-er_invalidColors)<< 4);
      }
      sumRot += rot;
      cubesBits |= 1 << cubes[cube];
   }

   if (cubesBits != m_cubesBits)
      // error, bad cubies colors;
      return 3 | ((-er_badColor)<< 4);

   else if (m_testOrientation && sumRot % m_nbc != 0)
      // error, cubies orientations not compatible
      return 3 | ((-er_badRotation)<< 4);

   else if (m_testPermutation && m_colorsCoding)
      // no error, parity out 1 or 2
      return 2 - PermutationsParity(cubes, maxCubies);

   else
      // no error, odd and even parity
      return 3;
}

//=========================================================================//
//                 initialize the computed solutions                       //
int CCompletion::CSlv::InitNodeTbl()                                       //
//=========================================================================//
{
   BYTE side;

   // initialize the solver tree
   for (m_pEnd = m_solvTbl, side = 0; side < 24; ++side) {
      if (m_work[side] < 6) continue;
      else if (m_pEnd >= m_solvTbl + maxNodeTbl) {
         // too many facets undefined
         TRACE("CCompletion::CSlv::InitNodeTbl_%d tooColors\n", m_nbc);
         return er_tooColors;
      }
      m_pEnd->side = side;
      ++m_pEnd;
   }
   return 0;
}

//=========================================================================//
//                      copy the computed solution                         //
void CCompletion::CSlv::copySolution(BYTE parity)                          //
//=========================================================================//
{
   // for computed sides
   for (Node *pCur = m_solvTbl; pCur < m_pEnd; ++pCur) {
      BYTE side = pCur->side;

      ASSERT(side < 24 && m_work[side] < 6);

      if (parity & 1) m_solutions[0][side] |= 1 << m_work[side];
      if (parity & 2) m_solutions[1][side] |= 1 << m_work[side];
   }

   ++m_sCnt;
   m_flags |= parity;
}

//=========================================================================//
//                                                                         //
//                      Calculate all possible solutions                   //
//                                                                         //
int CCompletion::CSlv::Computs()                                           //
//=========================================================================//
{
   memset(m_solutions, 0, 48);

   if ( (m_sCnt = InitNodeTbl()) != 0 ) {
      // too many facets undefined
      m_computed = false;
      m_flags = 3;               // error => movements even and odd

   } else if (m_pEnd == m_solvTbl) {
      // all facets are already defined, the test solution
      m_sCnt = ((m_flags = ValidityCheck())< 4)?
                         er_complete: -(m_flags>>4);
      m_computed = true;

   } else {
      Node *pCur = m_solvTbl;
      UINT count = 0;
      m_flags = 0;

      // for all possible solutions
     m_computed = true;
     while (pCur >= m_solvTbl) {

         if (pCur != m_pEnd) {
            // reads the parameters of the new step
            BYTE side = pCur->side;
            t_COLORS colors = (m_work[side] == 0xff)? getFreeColors(side):
                                                         pCur->colors;
            if (colors) {
               // update the current step and move to the next
               pCur->colors = colors ^ (1 <<(m_work[side]= colorCvrt(colors)));
               ++pCur;

            } else {
               // colors are exhausted, take a step back
               m_work[side] = 0xff;
               --pCur;
            }
         } else if ( ++count > maxTests ) {
            // exit, too long !

            // restore the work buffer
            for (pCur = m_solvTbl; pCur < m_pEnd; ++pCur)
               m_work[pCur->side] = 0xff;

            m_computed = false;
            m_sCnt = er_tooColors;
            m_flags = 3;         // movements even and odd
            break;

         } else {
            // all facets are defined, the solution is checked
            BYTE flags = ValidityCheck();

            //  save this good solution
            if (flags < 4) copySolution(flags);
            // to exhaust the other possibilities
            --pCur;
         }
      }

      if (m_sCnt == 0) {
         // no solution found
         m_flags = 3;
         m_sCnt = er_unsolvable;
      }
   }

   return m_sCnt;
}

//=========================================================================//
//                                                                         //
//                       Extract the calculated colors                     //
//                                                                         //
t_COLORS CCompletion::CSlv::getComputedColors( // return computed colors   //
                                    BYTE side,     // facet 0..23          //
                                 BYTE flags) const // permutations parity  //
//=========================================================================//
{
   t_COLORS colors = 0;

   ASSERT(side < 24);

   if (flags & 1) colors  = m_solutions[0][side];
   if (flags & 2) colors |= m_solutions[1][side];

   return colors;
}

//=========================================================================//
//                                                                         //
//                      Convert the output colors                          //
//                                                                         //
t_COLORS CCompletion::CSlv::colorsConvert(t_COLORS colors) const           //
//=========================================================================//
{
   t_COLORS cvrt;
   BYTE n;

   for (cvrt = n = 0, colors &= 0x3f; colors; ++n, colors >>= 1) {
      if (colors & 1) cvrt |= m_colorsId[n];
   }
   return cvrt;
}

//=========================================================================//
//                                                                         //
//                         Counts the color                                //
//                                                                         //
BYTE CCompletion::CSlv::getColorsCount(BYTE color) const                   //
//=========================================================================//
{
   BYTE count = 0;

   for (BYTE side = 0; side < 24; ++side) {
      if (m_work[side] == color) ++count;
   }
   return count;                 // for a color return 0 to 4
}

//=========================================================================//
//                                                                         //
//               Defines all the cubies completely filled                  //
//                                                                         //
t_CUBIES CCompletion::CSlv::getFilledCubes() const                         //
//=========================================================================//
{
   BYTE cur = 0, last, color;
   t_CUBIES used = 0;
   t_COLORS cvrt;

   do {
      last = cur + m_nbc, cvrt = 0;
      while ((color = m_work[cur]) < 6) {
         // make a cubie signature
         cvrt |= 1 << color;
         if (++cur == last) { used |= 1 << m_refCvrt[cvrt]; break; }
      }
   } while ((cur = last) < 24);

   return used;
}

//=========================================================================//
//                                                                         //
//       Test whether the proposed colors are valid for this corner        //
//                                                                         //
t_COLORS CCompletion::CSlv::checksCornerColor( // return valid color       //
                              BYTE side,  // facet of corner to fill 0..23 //
                        t_COLORS cn) const// color(s) available            //
//=========================================================================//
{
   BYTE c1, c2;      // no of the two colors

   switch (colorsCount(cn)) {
   case 1 :          // one box rest to fill

      c1 = colorCvrt(cn);

      if (CheckSidesColors(fillColor(side, c1))>= 0)
         return cvrtColor(c1);

      break;
   case 2 :          // two boxes are to be filled

      c1 = colorCvrt(cn);
      c2 = colorCvrt(cn ^ cvrtColor(c1));

      if (CheckSidesColors(FillColors(fillColor(side, c1), c2))>= 0)
         return cvrtColor(c1);
      if (CheckSidesColors(FillColors(fillColor(side, c2), c1))>= 0)
         return cvrtColor(c2);

      break;
   default :         // the three boxes are to be filled, the first color can
      return cn;     // be one of three
   }
   return 0;
}

//=========================================================================//
//                                                                         //
//          Test whether the proposed colors are valid for an edge         //
//                                                                         //
t_COLORS CCompletion::CSlv::searchEdgeColors(      // return valid colors  //
                  t_CUBIES used) const // pairs of colors already defined  //
//=========================================================================//
{
   BYTE color, cnt[6], ext[6], max, cube, n, t, sort[6], sum;
   t_COLORS cur, avail;

   // count the used colors
   for (sum = avail = color = 0; color < 6; ++color) {
      t = getColorsCount(color);
      cnt[color] = t;
      sum += t;
      if (t < 4) avail |= 1 << color;
   }

   if (sum < 12) {
      // sorting the used colors
      for (n = 0; n < 6; n++) sort[n] = n;
      do {
         for (t = 0xff, n = 0; n < 5; n++) {
            if (cnt[sort[n]] < cnt[sort[n+1]])
               t = sort[n], sort[n] = sort[n+1], sort[n+1] = t;
         }
      } while (t != 0xff);

      if (avail != 0x3f) {
         memcpy(ext, cnt, 6);

         // reserve the blocked colors
         for (n = 0; cnt[max = sort[n]] == 4; n++) {
            cur = 1 << max;
            for (color = 0; color < 6; ++color) {
               if ( max != color && (cube = m_refCvrt[cur|(1 << color)]) < 27
                               && (used & (1 << cube)) == 0 ) {
                  // reserve the color of the combination if it is not yet used
                  ++ ext[color];
               }
            }
         }

         // makes an inventory of available colors
         for (avail = color = 0; color < 6; ++color) {
            if (ext[color] < 4) avail |= 1 << color;
         }

         if (cnt[sort[2]] >= 3)
            avail &= ~((1 << sort[1]) | (1 << sort[2]));

      } else {

         if (cnt[sort[2]] >= 3)
            avail &= ~((1 << sort[0]) | (1 << sort[1]) | (1 << sort[2]));
      }

   }
/*
   TRACE("CSlv col[%d,%d,%d,%d,%d,%d] cnt[%d,%d,%d,%d,%d,%d] ext[%d,%d,%d,%d,%d,%d]\n",
      sort[0],sort[1],sort[2],sort[3],sort[4],sort[5],
      cnt[sort[0]],cnt[sort[1]],cnt[sort[2]],cnt[sort[3]],cnt[sort[4]],cnt[sort[5]],
      ext[sort[0]],ext[sort[1]],ext[sort[2]],ext[sort[3]],ext[sort[4]],ext[sort[5]]);
*/

   return avail;
}

//=========================================================================//
//                                                                         //
//                Test whether the proposed colors are valid               //
//                                                                         //
t_COLORS CCompletion::CSlv::searchColors() const   // return valid color   //
//=========================================================================//
{
   t_COLORS avail = 0;

   for (BYTE color = 0; color < 6; ++color) {
      if (getColorsCount(color) < 4) avail |= 1 << color;
   }
   return avail;
}

//=========================================================================//
//                                                                         //
//                 Gives the possible colors for this facet                //
//                                                                         //
t_COLORS CCompletion::CSlv::getFreeColors(BYTE side) const  // facet 0..23 //
//=========================================================================//
{
   t_COLORS colors = getColors(m_orgSides[side]) & 0x3f, avail;
   t_CUBIES used = getFilledCubes();

//   TRACE("CCompletion::CSlv::getFreeColors s=%d cs=%02x, u=%04x\n", side, colors, used);

   if (colors) avail = searchColors();
   // The cubie belongs to this facet has no defined colors :
   // Available colors are certainly usable ! This assertion is true only if
   // some structures are defined for the corners, but for the edges...
   else return (m_nbc == 2)? searchEdgeColors(used): searchColors();

   // if the structure is partially filled, calculates the missing colors
   t_COLORS out = 0, rc, cn;

   // examines all free colors signatures of the defined type
   const t_COLORS *pRefColors = m_refColors;
   while (rc = *pRefColors++) {
      // test whether the proposed colors are all contained in the signature
      // test if required colors are all available
      // test if the new signature color obtained, does not already been used
      cn = rc & ~colors;             // couleurs encore necessaires
      if ( colors == (rc & colors) && cn == (avail & cn)
                        && (used & (1 << m_refCvrt[rc]))== 0 ) {

         // removes the illegal arrangements of the corners
         out |= (m_nbc == 3)? checksCornerColor(side, cn): cn;
      }
   }

   return out;
}

//=========================================================================//
//                                                                         //
//                  Gives the possible colors for this facet               //
//                                                                         //
t_COLORS CCompletion::CSlv::getPossibleColors(                             //
                              BYTE facet, // requested facet number 0..54  //
                        BYTE flags) const // permutations parity of others //
//=========================================================================//
{
   BYTE side = m_reversSides[facet];   // translate to 0..23

   ASSERT(side < 24);
   if (side >= 24 || m_work[side] < 6) {
      return 0;
   }
   // selected the calculation program according to the state of the Cube
   t_COLORS colors = (m_computed)? getComputedColors(side, flags):
                                       getFreeColors(side);

   return colorsConvert(colors);
}

//=========================================================================//
//                                                                         //
//                  Update the facets of the Rubik's Cube                  //
//                                                                         //
BYTE CCompletion::CSlv::Update(BYTE facets[54], // Facet table to update   //
                           BYTE flags)          // permutations parity     //
//=========================================================================//
{
   BYTE side, count = 0;

   ASSERT(Compare(facets));

   for (side = 0; side < 24; ++side) {
      if ( m_work[side] < 6 ) continue;

      // only updates the free facets
      t_COLORS colors = (m_computed)? getComputedColors(side, flags):
                                          getFreeColors(side);

      if ( colors && colorsCount(colors) == 1 ) {
         // with one possible color
         facets[ m_convSides[side] ] = m_colorId[colorCvrt(colors)];
         ++count;
      }
   }
   if (count > 0) Create(facets);

   return count;
}

//=========================================================================//
//                                                                         //
int CCompletion::CSlv::getStatus(BYTE flags) const                         //
//=========================================================================//
{
   return (flags & m_flags)? m_sCnt: er_badPermut;
}

//=========================================================================//
//                                                                         //
//                 Computs the error or completion message                 //
//                                                                         //
CString CCompletion::CSlv::getErrorString(int err, BYTE flags) const      //
//=========================================================================//
{
   static const TCHAR *obj[2] = {"Artes", "Sommets"};
   CString msg;

   if (err == 0 ) {
      return "";
   } else if (err == 1) {
      msg = " 1 possibilit";
   } else if (err > 2 && m_flags == 3 && flags != 3) {
      msg.Format(" %d possibilits", err / 2);
   } else if (err > 1) {
      msg.Format(" %d possibilits", err);
   } else if (err <= er_last ) {
      msg.Format(": erreur_%d", -err);
   } else {
      static const TCHAR *errorMsg[] = {
         "",
         " pas de donnes disponibles",                              // -1
         " c'est complet !",                                         // -2
         "  complter...",                                          // -3
         " mauvaises couleurs !",                                    // -4
         " configuration couleurs incompatible !",                   // -5
         " configuration insoluble !",                               // -6
         " rotations incompatibles !",                               // -7
         "Les couleurs des centres de faces doivent tre dfinies",  // -8
         "Les permutations des couleurs sont incompatibles !"        // -9
      };
   // er_missData=-1, er_complete=-2, er_missColor=-3, er_badColor=-4,
   // er_invalidColors=-5, er_unsolvable=-6, er_badRotation=-7,
   // er_centersUndefine=-8, er_badPermut=-9, er_last=-10
      msg = errorMsg[-err];
   }

   if (msg.GetAt(0) == TCHAR(' '))
      msg.Insert(0, obj[m_nbc - 2]);

   return msg + "\r\n";
}

//=========================================================================//
CString CCompletion::CSlv::Dump(const BYTE wrk[24]) const
//=========================================================================//
{
   CString out, tmp;

   for (BYTE n = 0; n < 24; n++) {
      tmp.Format(" %02X", wrk[n]);
      out += tmp;
   }
   return out;
}

/////////////////////////////////////////////////////////////////////////////